@@ -32,23 +32,30 @@ package Geo::Coder::Free::DB;
 # my $row = $foo->fetchrow_hashref(customer_id => '12345);
 # print Data::Dumper->new([$row])->Dump();
 
+# CSV files can have empty lines of comment lines starting with '#', to make them more readable
+
+# If the table has a column called "entry", sorts are based on that
+# To turn that off, pass 'no_entry' to the constructor, for legacy
+# reasons it's enabled by default
+# TODO: Switch that to off by default, and enable by passing 'entry'
+
 # TODO: support a directory hierachy of databases
 # TODO: consider returning an object or array of objects, rather than hashes
+# TODO:	Add redis database - could be of use for Geo::Coder::Free
+#	use select() to select a database - use the table arg
+#	new(database => 'redis://servername');
 
 use warnings;
 use strict;
 
+use DBD::SQLite::Constants qw/:file_open/;	# For SQLITE_OPEN_READONLY
 use File::Basename;
-use DBI;
 use File::Spec;
 use File::pfopen 0.02;
 use File::Temp;
-use Gzip::Faster;
-use DBD::SQLite::Constants qw/:file_open/;	# For SQLITE_OPEN_READONLY
 use Error::Simple;
 use Carp;
 
-our @databases;
 our $directory;
 our $logger;
 our $cache;
@@ -70,7 +77,8 @@ sub new {
 		logger => $args{'logger'} || $logger,
 		directory => $args{'directory'} || $directory,	# The directory conainting the tables in XML, SQLite or CSV format
 		cache => $args{'cache'} || $cache,
-		table => $args{'table'}	# The name of the file containing the table, defaults to the class name
+		table => $args{'table'},	# The name of the file containing the table, defaults to the class name
+		no_entry => $args{'no_entry'} || 0,
 	}, $class;
 }
 
@@ -81,9 +89,6 @@ sub init {
 	$directory ||= $args{'directory'};
 	$logger ||= $args{'logger'};
 	$cache ||= $args{'cache'};
-	if($args{'databases'}) {
-		@databases = $args{'databases'};
-	}
 }
 
 sub set_logger {
@@ -104,6 +109,8 @@ sub set_logger {
 	$self->{'logger'} = $args{'logger'};
 }
 
+# Open the database.
+
 sub _open {
 	my $self = shift;
 	my %args = (
@@ -129,6 +136,10 @@ sub _open {
 	}
 
 	if(-r $slurp_file) {
+		require DBI;
+
+		DBI->import();
+
 		$dbh = DBI->connect("dbi:SQLite:dbname=$slurp_file", undef, undef, {
 			sqlite_open_flags => SQLITE_OPEN_READONLY,
 		});
@@ -137,10 +148,14 @@ sub _open {
 		if($self->{'logger'}) {
 			$self->{'logger'}->debug("read in $table from SQLite $slurp_file");
 		}
+		$self->{'type'} = 'DBI';
 	} else {
 		my $fin;
 		($fin, $slurp_file) = File::pfopen::pfopen($dir, $table, 'csv.gz:db.gz');
 		if(defined($slurp_file) && (-r $slurp_file)) {
+			require Gzip::Faster;
+			Gzip::Faster->import();
+
 			close($fin);
 			$fin = File::Temp->new(SUFFIX => '.csv', UNLINK => 0);
 			print $fin gunzip_file($slurp_file);
@@ -185,7 +200,11 @@ sub _open {
 				f_file => $slurp_file,
 				escape_char => '\\',
 				sep_char => $sep_char,
-				auto_diag => 1,
+				# Don't do this, causes "Bizarre copy of HASH
+				#	in scalar assignment in error_diag
+				#	RT121127
+				# auto_diag => 1,
+				auto_diag => 0,
 				# Don't do this, it causes "Attempt to free unreferenced scalar"
 				# callbacks => {
 					# after_parse => sub {
@@ -237,7 +256,9 @@ sub _open {
 			)};
 
 			# Don't use blank lines or comments
-			@data = grep { $_->{'entry'} !~ /^#/ } grep { defined($_->{'entry'}) } @data;
+			unless($self->{no_entry}) {
+				@data = grep { $_->{'entry'} !~ /^\s*#/ } grep { defined($_->{'entry'}) } @data;
+			}
 			# $self->{'data'} = @data;
 			my $i = 0;
 			$self->{'data'} = ();
@@ -245,6 +266,7 @@ sub _open {
 				$self->{'data'}[$i++] = $d;
 			}
 			}
+			$self->{'type'} = 'CSV';
 		} else {
 			$slurp_file = File::Spec->catfile($dir, "$table.xml");
 			if(-r $slurp_file) {
@@ -257,11 +279,10 @@ sub _open {
 			} else {
 				throw Error::Simple("Can't open $dir/$table");
 			}
+			$self->{'type'} = 'XML';
 		}
 	}
 
-	push @databases, $table;
-
 	$self->{$table} = $dbh;
 	my @statb = stat($slurp_file);
 	$self->{'_updated'} = $statb[9];
@@ -289,19 +310,55 @@ sub selectall_hash {
 		if($self->{'logger'}) {
 			$self->{'logger'}->trace("$table: selectall_hash fast track return");
 		}
-		return @{$self->{'data'}};
+		# This use of a temporary variable is to avoid
+		#	"Implicit scalar context for array in return"
+		# return @{$self->{'data'}};
+		my @rc = @{$self->{'data'}};
+		return @rc;
 	}
+	# if((scalar(keys %params) == 1) && $self->{'data'} && defined($params{'entry'})) {
+	# }
 
-	my $query = "SELECT * FROM $table";
+	my $query;
+	my $done_where = 0;
+	if(($self->{'type'} eq 'CSV') && !$self->{no_entry}) {
+		$query = "SELECT * FROM $table WHERE entry IS NOT NULL AND entry NOT LIKE '#%'";
+		$done_where = 1;
+	} else {
+		$query = "SELECT * FROM $table";
+	}
 	my @query_args;
 	foreach my $c1(sort keys(%params)) {	# sort so that the key is always the same
-		if(scalar(@query_args) == 0) {
-			$query .= ' WHERE';
-		} else {
-			$query .= ' AND';
+		my $arg = $params{$c1};
+		if(ref($arg)) {
+			if($self->{'logger'}) {
+				$self->{'logger'}->fatal("selectall_hash $query: argument is not a string");
+			}
+			throw Error::Simple("$query: argument is not a string");
 		}
-		$query .= " $c1 = ?";
-		push @query_args, $params{$c1};
+		if(!defined($arg)) {
+			my @call_details = caller(0);
+			throw Error::Simple("$query: value for $c1 is not defined in call from " .
+				$call_details[2] . ' of ' . $call_details[1]);
+		}
+		if($done_where) {
+			if($arg =~ /\@/) {
+				$query .= " AND $c1 LIKE ?";
+			} else {
+				$query .= " AND $c1 = ?";
+			}
+		} else {
+			if($arg =~ /\@/) {
+				$query .= " WHERE $c1 LIKE ?";
+			} else {
+				$query .= " WHERE $c1 = ?";
+			}
+			$done_where = 1;
+		}
+		push @query_args, $arg;
+	}
+	if(!$self->{no_entry}) {
+		$query .= ' ORDER BY entry';
 	}
 	if($self->{'logger'}) {
 		if(defined($query_args[0])) {
@@ -317,12 +374,17 @@ sub selectall_hash {
 	my $c;
 	if($c = $self->{cache}) {
 		if(my $rc = $c->get($key)) {
-			return @{$rc};
+			# This use of a temporary variable is to avoid
+			#	"Implicit scalar context for array in return"
+			# return @{$rc};
+			my @rc = @{$rc};
+			return @rc;
 		}
 	}
 
 	if(my $sth = $self->{$table}->prepare($query)) {
-		$sth->execute(@query_args) || throw Error::Simple("$query: @query_args");
+		$sth->execute(@query_args) ||
+			throw Error::Simple("$query: @query_args");
 
 		my @rc;
 		while(my $href = $sth->fetchrow_hashref()) {
@@ -335,7 +397,9 @@ sub selectall_hash {
 
 		return @rc;
 	}
-	$self->{'logger'}->warn("selectall_hash failure on $query: @query_args");
+	if($self->{'logger'}) {
+		$self->{'logger'}->warn("selectall_hash failure on $query: @query_args");
+	}
 	throw Error::Simple("$query: @query_args");
 }
 
@@ -357,26 +421,48 @@ sub fetchrow_hashref {
 	} else {
 		$query .= $table;
 	}
-	my @args;
+	my $done_where = 0;
+	if(($self->{'type'} eq 'CSV') && !$self->{no_entry}) {
+		$query .= " WHERE entry IS NOT NULL AND entry NOT LIKE '#%'";
+		$done_where = 1;
+	}
+	my @query_args;
 	foreach my $c1(sort keys(%params)) {	# sort so that the key is always the same
-		if(scalar(@args) == 0) {
-			$query .= ' WHERE';
-		} else {
-			$query .= ' AND';
+		if(my $arg = $params{$c1}) {
+			if($done_where) {
+				if($arg =~ /\@/) {
+					$query .= " AND $c1 LIKE ?";
+				} else {
+					$query .= " AND $c1 = ?";
+				}
+			} else {
+				if($arg =~ /\@/) {
+					$query .= " WHERE $c1 LIKE ?";
+				} else {
+					$query .= " WHERE $c1 = ?";
+				}
+				$done_where = 1;
+			}
+			push @query_args, $arg;
 		}
-		$query .= " $c1 = ?";
-		push @args, $params{$c1};
 	}
 	# $query .= ' ORDER BY entry LIMIT 1';
 	$query .= ' LIMIT 1';
 	if($self->{'logger'}) {
-		if(defined($args[0])) {
-			$self->{'logger'}->debug("fetchrow_hashref $query: ", join(', ', @args));
+		my @call_details = caller(0);
+		if(defined($query_args[0])) {
+			$self->{'logger'}->debug("fetchrow_hashref $query: ", join(', ', @query_args),
+				' called from ', $call_details[2] . ' of ' . $call_details[1]);
 		} else {
 			$self->{'logger'}->debug("fetchrow_hashref $query");
 		}
 	}
-	my $key = "fetchrow $query " . join(', ', @args);
+	my $key;
+	if(defined($query_args[0])) {
+		$key = "fetchrow $query " . join(', ', @query_args);
+	} else {
+		$key = "fetchrow $query";
+	}
 	my $c;
 	if($c = $self->{cache}) {
 		if(my $rc = $c->get($key)) {
@@ -384,7 +470,7 @@ sub fetchrow_hashref {
 		}
 	}
 	my $sth = $self->{$table}->prepare($query) or die $self->{$table}->errstr();
-	$sth->execute(@args) || throw Error::Simple("$query: @args");
+	$sth->execute(@query_args) || throw Error::Simple("$query: @query_args");
 	if($c) {
 		my $rc = $sth->fetchrow_hashref();
 		$c->set($key, $rc, '1 hour');
@@ -461,24 +547,37 @@ sub AUTOLOAD {
 	my %params = (ref($_[0]) eq 'HASH') ? %{$_[0]} : @_;
 
 	my $query;
+	my $done_where = 0;
 	if(wantarray && !delete($params{'distinct'})) {
-		$query = "SELECT $column FROM $table";
+		if(($self->{'type'} eq 'CSV') && !$self->{no_entry}) {
+			$query = "SELECT $column FROM $table WHERE entry IS NOT NULL AND entry NOT LIKE '#%'";
+			$done_where = 1;
+		} else {
+			$query = "SELECT $column FROM $table";
+		}
 	} else {
-		$query = "SELECT DISTINCT $column FROM $table";
+		if(($self->{'type'} eq 'CSV') && !$self->{no_entry}) {
+			$query = "SELECT DISTINCT $column FROM $table WHERE entry IS NOT NULL AND entry NOT LIKE '#%'";
+			$done_where = 1;
+		} else {
+			$query = "SELECT DISTINCT $column FROM $table";
+		}
 	}
 	my @args;
-	foreach my $c1(keys(%params)) {
-		if(!defined($params{$c1})) {
-			$self->{'logger'}->debug("AUTOLOAD params $c1 isn't defined");
-		}
-		# $query .= " AND $c1 LIKE ?";
-		if(scalar(@args) == 0) {
-			$query .= ' WHERE';
+	while(my ($key, $value) = each %params) {
+		if(defined($value)) {
+			if($done_where) {
+				$query .= " AND $key = ?";
+			} else {
+				$query .= " WHERE $key = ?";
+				$done_where = 1;
+			}
+			push @args, $value;
 		} else {
-			$query .= ' AND';
+			if($self->{'logger'}) {
+				$self->{'logger'}->debug("AUTOLOAD params $key isn't defined");
+			}
 		}
-		$query .= " $c1 = ?";
-		push @args, $params{$c1};
 	}
 	$query .= " ORDER BY $column";
 	if(!wantarray) {
